# Generated file
# - do not overwrite
# - do not include in git
from collections import namedtuple

BuildInfo = namedtuple(
    'BuildInfo',
    ['version', 'built_at', 'sha', 'branch', 'tag'],
)

BUILD_INFO = BuildInfo(
    version='v3.24.0~de1a777',
    built_at='2023-05-31 09:40:30Z',
    sha='de1a777fb9287f666e161e98200054ae757fa44c',
    branch='HEAD',
    tag='v3.24.0',
)
